/*************************************************************************
 * *** This is NeTAMS, version 4
 * ***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
 * ***	(c) 2002-2006 NeTAMS Development Team
 * ***	All rights reserved. See 'Copying' file included in distribution
 * ***	For latest version and more info, visit this project web page
 * ***	located at http://www.netams.com
 * ***
 * *************************************************************************/
/* $Id: config.h 1628 2013-10-28 19:08:22Z anton $ */

#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <time.h>
#ifndef WIN32
#define __FAVOR_BSD
#include <sys/socket.h>
#include <dlfcn.h>
#include <netinet/in_systm.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <net/ethernet.h>
#include <arpa/inet.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <poll.h>
#include <sys/utsname.h>
#include <net/if.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#ifdef LINUX
#include <netinet/ether.h>
#endif
#endif
#include <sys/types.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <syslog.h>
#include <libgen.h>

#include "lib.h"

void n4debug(const char* fmt, ...);

#ifdef WIN32
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

class flow;
typedef void (submit_data_handler)(flow *flow_to_send);

struct agent_config {
	struct timeval program_start, program_end;
	time_t start_t, end_t;
	int run; // goes to 0 if we have to quit
	const char *password;
	const char *connect_string;
	int quiet;
	unsigned long msgs_in, msgs_out, bad_pkts;
	unsigned long seq;

	int sockfd;
	struct sockaddr_in servaddr;
	int pktlen;
	unsigned char *sendbuffer;
	int data_lastpos;
	unsigned char *data_sendbuffer;
	int auth; // 1=flow auth required
	char *exec_cmd;
	char *config_file_name;
	char *pid_file_name;
	u_char daemon;
};

#include "ds_libpcap.h"
#include "flow.h"

#define IPP_DUMMY 0
#define IPP_FRAMES 1
#define IPP_OCTETS 2
#define IPP_IPSRC 3
#define IPP_IPDST 4
#define IPP_IPPROTO 5
#define IPP_IPTOS 6
#define IPP_SRCPORT 7
#define IPP_DSTPORT 8
#define IPP_URL 9
#define IPP_DPI 10

#define IPP_NEXTFOLLOW 128
// #define IPP_NONEXT 129
#define IPP_SEQ 130
#define IPP_MAX_PKT_LEN 1450

#define PERMIT 		1
#define BLOCK  		2
#define NOTFOUND	3
#define LOGIN		4
#define LOGOUT		5
#define RESTRICT	6
#define FLUSHALL	7

extern const char *action_name[];

#define MSG_REGISTER	1
#define MSG_SETFILTER	2
#define MSG_STATUS	4
#define MSG_AUTHACTION  5

extern const char *msgtype_name[];

extern unsigned long long	mem_total_alloc;
extern unsigned long long	mem_url_alloc;
extern unsigned long long	mem_total_free;
extern unsigned long long	mem_url_free;

extern unsigned long long	mem_c_total_alloc;
extern unsigned long long	mem_c_url_alloc;
extern unsigned long long	mem_c_total_free;
extern unsigned long long	mem_c_url_free;

//lib.c
void setULong(unsigned char data[], int offset, long value);
long getULong(unsigned char data[], int offset);
void setUShort(unsigned char data[], int offset, int value);
int getUShort(unsigned char data[], int offset);
void setString(unsigned char data[], int offset, const char* value, int length);
char *set_string(const char *string);

// maximum length of stored URL (layer7 filter)
#define MAX_URL 58
int layer7_checkinfo(u_int16_t port, flow *flow_entry, struct ip *ip);

#ifdef ENABLE_DPI
#include "libopendpi-1.3/libopendpi/ipq_api.h"
extern u_int32_t size_id_struct;
extern u_int32_t size_flow_struct;
extern u_int32_t detection_tick_resolution;
extern char *prot_short_str[];

void init_dpi();
void dpi(flow* flow_entry, struct ip *ip, int length, u_int64_t time);
#endif
