/*************************************************************************
***	Authentication, authorization, accounting + firewalling package
***	Copyright 1998-2002 Anton Vinokurov <anton@netams.com>
***	Copyright 2002-2008 NeTAMS Development Team
***	This code is GPL v3
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: flow.h 1628 2013-10-28 19:08:22Z anton $ */

#ifndef NA_FLOW_H
#define NA_FLOW_H

#define ATTR_FLOW_INFO	0
struct	flow_info_value {
	u_long		packets;
	u_int64_t	octets;
	time_t		flow_first;
	time_t		flow_last;
	u_char		direction;
} __attribute__((packed));

#define	ATTR_IPV4_INFO	1
struct	ipv4_info_value {
    struct in_addr	ip_src, ip_dst;
    u_char		ip_p;
    u_char		ip_tos;
} __attribute__((packed));

#define ATTR_TCP_INFO	2
struct tcp_info_value {
    u_short	src_port;
    u_short	dst_port;
} __attribute__((packed));

#define	ATTR_MAC_INFO	3
struct mac_info_value {
	u_char		src[ETHER_ADDR_LEN];
	u_char		dst[ETHER_ADDR_LEN];
} __attribute__((packed));

#define	ATTR_VLAN_INFO	4
#define DOT1X_VLAN_MASK	0x0FFF
#define DOT1X_GET_VLAN(value) ((value) & DOT1X_VLAN_MASK)
#define DOT1X_GET_CFI(value) (((value) >> 12) & 1)
#define DOT1X_GET_PRI(value) (((value) >> 13) & 7)
struct vlan_info_value {
	u_int16_t	dot1x;
} __attribute__((packed));

#define	ATTR_AS_INFO	5
struct	as_info_value {
    u_short	as_src;
    u_short	as_dst;
} __attribute__((packed));

#define	ATTR_IFINDEX_INFO	6
struct	ifindex_info_value {
    u_short	if_in;
    u_short	if_out;
} __attribute__((packed));

#define ATTR_MULTICAST_INFO	7
struct multicast_info_value {
    u_int64_t	dst_packets;
    u_int64_t	dst_bytes;
    u_char	igmp_type;
} __attribute__((packed));

#define ATTR_NEXTHOP_INFO	8
struct nexthop_info_value {
    struct in_addr	ipv4_nexthop;
    struct in_addr	bgpv4_nexthop;
} __attribute__((packed));

#define ATTR_URL_INFO	9
struct url_info_value {
    u_char len;
    char* value;
} __attribute__((packed));


struct ipv4_key {
	struct ipv4_info_value	ipv4_info;
	struct tcp_info_value	tcp_info;
	u_char			padding[2];
};

struct pcap_ipv4_key {
	struct ipv4_info_value	ipv4_info;
	struct tcp_info_value	tcp_info;
	u_short			dot1x;
};
//////////////////////////////////////////////////////////////////////////
struct ipv4_key* IPv4GetKey(struct ip *ip, struct ipv4_key *key);
//void IPv4FillFlow(struct ipv4_key* ipv4_flow_key, entry *flow_entry);

union attribute_value {
	struct flow_info_value		flow_info;
	struct ipv4_info_value		ipv4_info;
	struct tcp_info_value		tcp_info;
	struct mac_info_value		mac_info;
	struct vlan_info_value		vlan_info;
	struct as_info_value		as_info;
	struct ifindex_info_value	ifindex_info;
	struct multicast_info_value	multicast_info;
	struct nexthop_info_value	nexthop_info;
	struct url_info_value		url_info;
} __attribute__((packed));

struct	attribute {
	u_short			id;
	union attribute_value	value;
} __attribute__((packed));

//////////////////////////////////////////////////////////////////////////
#define FLOW_CACHE_SIZE 0x10000
#define FLOW_CACHE_MAXCHUNK 2048
#define FLOW_HASH(idx)  CRC16(idx, 16);
#define FLOW_PASS_DEFAULT	0
#define FLOW_PASS_ALLOW		1
#define FLOW_PASS_DENY		2
#define FLOW_FLUSH			4
#define FLOW_EXPIRETIME		30

#define IPP_MAX_DPI_LEN 16
#define FLOW_SHORT_HASH(idx)  (idx->ipv4_info.ip_src.s_addr + idx->ipv4_info.ip_dst.s_addr + idx->ipv4_info.ip_p)


class flow {
	public:
		flow();
		~flow();

		u_int16_t	hash;
		struct ipv4_key	key;
		flow		*next;
		flow		*ex_next; //for expiresearch;
		char		authflags;
		char		flushflag;
		unsigned	agg_count;
		unsigned long	ex_packets;
		time_t		ex_time;
		time_t		st_time;
		unsigned long	packets;
		unsigned long long	bytes;
		unsigned	status;
#ifdef URL_CHECK
		unsigned 	url_len;
		char*		url_value;	
#endif

#ifdef ENABLE_DPI
		u_int32_t	dpi_protocol;
		struct ipoque_flow_struct *ipoque_flow;
#endif

};
#define FLOW_STATUS_NEW	1
#define FLOW_STATUS_AUTH_SEND	2
#define FLOW_STATUS_AUTH_TIMEOUT	3
#define FLOW_STATUS_AUTH_SUCCESS	4
#define FLOW_STATUS_AUTH_DENIED	5
#define FLOW_STATUS_EXPIRED	6

#define PURGE_THRESHOLD	128*1024 /* flows greater than 128K are routed straight, never cached */

//////////////////////////////////////////////////////////////////////////
class flowcache {
	public:
		flowcache(struct ds_pcap_info *pcap_info);

		struct ds_pcap_info *pcap_info;
		flow	**cache;
		flow	*ex_root; // for expiresearch
		flow	*ex_last; // for expiresearch
		flow	*match(struct ipv4_key *key, unsigned long packets, unsigned long long bytes);
		void 	expire();
		void 	dump();
		void 	flush();
		flow	*find_rev_flow(flow *original);
		unsigned	entries;
		unsigned long long	total_entries;
		unsigned long long	flushed_entries;
		unsigned long long	routed_entries;

};

//////////////////////////////////////////////////////////////////////////
#endif
