/*************************************************************************
*** This is NeTAMS, version 4
***	(c) 1998-2002 Anton Vinokurov <anton@netams.com>
***	(c) 2002-2010 NeTAMS Development Team
***	All rights reserved. See 'Copying' file included in distribution
***	For latest version and more info, visit this project web page
***	located at http://www.netams.com
***
*************************************************************************/
/* $Id: n4agent_pcap_event.c 1628 2013-10-28 19:08:22Z anton $ */

#include "config.h"
#include <sys/time.h>
#include <event.h>

#if defined MACOS
#define EVENT_NOPOLL 
#define EVENT_NOKQUEUE
#endif

/* global starter variables */
const char *revision="$Rev: 1628 $";
FILE *configfile;
char *pid_file_name=NULL;
char *exec_cmd_name=NULL;
u_char is_daemon=0,is_quiet=0,optimization=2;

unsigned long long	mem_total_alloc=0;
unsigned long long	mem_url_alloc=0;
unsigned long long	mem_total_free=0;
unsigned long long	mem_url_free=0;

unsigned long long	mem_c_total_alloc=0;
unsigned long long	mem_c_url_alloc=0;
unsigned long long	mem_c_total_free=0;
unsigned long long	mem_c_url_free=0;

struct agent_config *ac=NULL;
struct ds_pcap_info *pcap_info;
struct timeval expire_tv;

char *destination="127.0.0.1:20002";
char *rule="ip";
char *interface="lo";
char *password="password";
unsigned active=120;
unsigned fc_period=10;

struct clientfd {
    int fd;
    struct event ev;
} client[4];

//FILE* for pipes
FILE* exec=NULL;

void run_exec();

void server_callback (int fd, short event, void *arg);
void pcap_callback (int fd, short event, void *arg);
void stdin_callback (int fd, short event, void *arg);
void flow_expire_callback (int fd, short event, void *arg);

void signal_handler(int signal){
	n4debug("SIGNAL = %d\n", signal);
	if (ac!=NULL) if (signal==SIGHUP || signal==SIGTERM || signal==SIGINT) ac->run=0;
	if (signal==SIGUSR1)  {
		if (pcap_info->is_registered) {
			syslog(LOG_INFO, "NETAMS4 remote agent is registered, in=%lu, out=%lu, bad=%lu, fc_size=%u\n", ac->msgs_in, ac->msgs_out, ac->bad_pkts, pcap_info->fc->entries);
		    struct pcap_stat ps;
			pcap_stats(pcap_info->pcap, &ps);
			syslog(LOG_INFO, "flowcache stats: total_entries=%llu, flushed_entries=%llu, routed_entries=%llu\n", pcap_info->fc->total_entries, pcap_info->fc->flushed_entries, pcap_info->fc->routed_entries);
			syslog(LOG_INFO, "mem stats: mem_total_alloc=%llu, mem_url_alloc=%llu, mem_total_free=%llu, mem_url_free=%llu\n", mem_total_alloc, mem_url_alloc, mem_total_free, mem_url_free);
			syslog(LOG_INFO, "mem stats: mem_c_total_alloc=%llu, mem_c_url_alloc=%llu, mem_c_total_free=%llu, mem_c_url_free=%llu\n", mem_c_total_alloc, mem_c_url_alloc, mem_c_total_free, mem_c_url_free);
			syslog(LOG_INFO, "pcap stats: pkt_recv=%u, pkt_drop=%u, pkt_ifdrop=%u\n", ps.ps_recv, ps.ps_drop, ps.ps_ifdrop);
		}
		else syslog(LOG_INFO, "NETAMS4 remote agent is not registered with billing core data source");
	}
}

unsigned long crc32p(const unsigned char *buf, const char *password, unsigned int len);
unsigned long gethash(unsigned char *buf, const char *password, unsigned int len){
	unsigned long i=getULong(buf, 3);
	setULong(buf, 3, 0);
	unsigned long j=crc32p(buf, password, len);
	setULong(buf, 3, i);
	return j;
}

void submit_registration() {
	ac->pktlen=512;
	memset(ac->sendbuffer, '\0', ac->pktlen);
	ac->sendbuffer[0]=2; // version
	setUShort(ac->sendbuffer, 1, ac->pktlen); // length
	ac->sendbuffer[7]=1; // MSG_REGISTER
	setUShort(ac->sendbuffer, 8, (ac->exec_cmd!=NULL?1:0)); // can we block or not?
	const char rev[] = "$Rev: 1628 $";
	char *id = (char*)malloc(32); snprintf(id, 32, "pcap_agent_ev %s", rev);
	setString(ac->sendbuffer, 10, id, 32);
	free(id);
	setULong(ac->sendbuffer, 3, crc32p(ac->sendbuffer, ac->password, ac->pktlen)); // sign & add hash
	sendto(ac->sockfd, ac->sendbuffer, ac->pktlen, 0, (struct sockaddr *)&(ac->servaddr), sizeof(struct sockaddr));
	ac->msgs_out++;
	n4debug("sending registration\n");
}

void submit_keepalive() {
	ac->pktlen=32;
	memset(ac->sendbuffer, '\0', ac->pktlen);
	ac->sendbuffer[0]=2; // version
	setUShort(ac->sendbuffer, 1, ac->pktlen); // length
	ac->sendbuffer[7]=4; // MSG_STATUS
	setUShort(ac->sendbuffer, 8, (ac->exec_cmd!=NULL?1:0)); // can we block or not?
	setULong(ac->sendbuffer, 10, pcap_info->fc->entries);
	setULong(ac->sendbuffer, 3, crc32p(ac->sendbuffer, ac->password, ac->pktlen)); // sign & add hash
	sendto(ac->sockfd, ac->sendbuffer, ac->pktlen, 0, (struct sockaddr *)&(ac->servaddr), sizeof(struct sockaddr));
	ac->msgs_out++;
	n4debug("sending keepalive\n");
}
void forward_data_packet() {
	ac->data_sendbuffer[ac->data_lastpos]=IPP_DUMMY;
	setUShort(ac->data_sendbuffer, 1, ac->data_lastpos); // length
	setULong(ac->data_sendbuffer, 3, crc32p(ac->data_sendbuffer, ac->password,  ac->data_lastpos)); // sign & add hash
	sendto(ac->sockfd, ac->data_sendbuffer, ac->data_lastpos, 0, (struct sockaddr *)&(ac->servaddr), sizeof(struct sockaddr));
	ac->msgs_out++;
	n4debug("send %u %lu\n", ac->data_lastpos, ac->seq);
	ac->data_lastpos=0;
}

void submit_data(flow *to_send) {
	if (ac->auth==0 && to_send->status==FLOW_STATUS_NEW) return; // no auth and new flow
	else if (ac->auth==1 && to_send->status==FLOW_STATUS_NEW) { // auth and new flow: send auth req
		to_send->status=FLOW_STATUS_AUTH_SEND;
		return; // TODO
	} else { // expired or similar
		char p1[32], p2[32];
		ipv4_key *k = &to_send->key;
		n4debug("route %p  [%u %s:%u -> %s:%u] %lu %llu\n", to_send, k->ipv4_info.ip_p, inet_ntop(AF_INET, &k->ipv4_info.ip_src, p1, 32), ntohs(k->tcp_info.src_port), inet_ntop(AF_INET, &k->ipv4_info.ip_dst, p2, 32), ntohs(k->tcp_info.dst_port), to_send->packets, to_send->bytes);
		if (ac->data_lastpos==0) { // init header
			memset(ac->data_sendbuffer, '\0', 1500);
			ac->data_sendbuffer[0]=2; // version
			ac->data_sendbuffer[7]=3; // MSG_DATA
			ac->data_sendbuffer[8]=1; // IP payload
			ac->data_lastpos=9;
		}

		ac->data_sendbuffer[ac->data_lastpos]=IPP_SEQ; setULong(ac->data_sendbuffer, ac->data_lastpos+1, ++ac->seq); ac->data_lastpos+=5;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_FRAMES; setUShort(ac->data_sendbuffer, ac->data_lastpos+1, to_send->packets); ac->data_lastpos+=3;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_OCTETS; setULong(ac->data_sendbuffer, ac->data_lastpos+1, to_send->bytes); ac->data_lastpos+=5;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_IPSRC; setULong(ac->data_sendbuffer, ac->data_lastpos+1, ntohl(k->ipv4_info.ip_src.s_addr));  ac->data_lastpos+=5;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_IPDST; setULong(ac->data_sendbuffer, ac->data_lastpos+1, ntohl(k->ipv4_info.ip_dst.s_addr)); ac->data_lastpos+=5;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_IPPROTO; setUShort(ac->data_sendbuffer, ac->data_lastpos+1, k->ipv4_info.ip_p);  ac->data_lastpos+=3;
		ac->data_sendbuffer[ac->data_lastpos]=IPP_IPTOS; ac->data_sendbuffer[ac->data_lastpos+1]=k->ipv4_info.ip_tos; ac->data_lastpos+=2;
		if (k->ipv4_info.ip_p==6 || k->ipv4_info.ip_p==17) {
			ac->data_sendbuffer[ac->data_lastpos]=IPP_SRCPORT; setUShort(ac->data_sendbuffer, ac->data_lastpos+1, ntohs(k->tcp_info.src_port)); ac->data_lastpos+=3;
			ac->data_sendbuffer[ac->data_lastpos]=IPP_DSTPORT; setUShort(ac->data_sendbuffer, ac->data_lastpos+1, ntohs(k->tcp_info.dst_port)); ac->data_lastpos+=3;
		}

		#ifdef URL_CHECK
		if (to_send->url_len>0 && to_send->url_value!=NULL){
			ac->data_sendbuffer[ac->data_lastpos]=IPP_URL;
			ac->data_sendbuffer[ac->data_lastpos+1]=(char)(to_send->url_len);
			memcpy(&ac->data_sendbuffer[ac->data_lastpos+2], to_send->url_value, to_send->url_len);
			ac->data_lastpos=ac->data_lastpos+2+ to_send->url_len;
			n4debug("\turl len=%d '%s' lastpos=%d\n",  to_send->url_len, to_send->url_value, ac->data_lastpos);
			}
		#endif

		#ifdef ENABLE_DPI
		n4debug("\tdpi protocol=%d '%s'\n",  to_send->dpi_protocol, prot_short_str[to_send->dpi_protocol]);
		#endif

		ac->data_sendbuffer[ac->data_lastpos]=IPP_NEXTFOLLOW;
		ac->data_lastpos+=1;

		if (ac->data_lastpos>IPP_MAX_PKT_LEN-64) forward_data_packet();

	}
}

void usage(char *param){
	printf("n4agent_pcap for NETAMS 4.0 billing system, %s \n", revision);
	printf("Usage: %s <parameters>\n", param);
	printf("\t-h print this parameter help info\n");
	printf("\t-q be quiet and don't output anything\n");
	printf("\t-d become a daemon (detach from terminal)\n");
	printf("\t-p PID file name\n");
	printf("\t-e command to execute on received status change (access-script)\n");
	printf("\t-b libpcap buffer size in megabytes (1 to 256)\n");
	printf("\t-c connect to jserver at \"host:port\", default is \"127.0.0.1:20002\"\n");
	printf("\t-r libpcap capture rule, default is \"ip\"\n");
	printf("\t-i libpcap capture interface\n");
	printf("\t-t active flow timeout, default is 120 seconds\n");
	printf("\t-s protect communication with jserver with password\n");
	printf("\t-o flow cache optimization level (default: 2)\n");
	printf("\t-x flow cache expirecheck periodicity (default: 10 seconds)\n");
	printf("\n");
	exit(-1);
}

int main(int argc, char *argv[]){
	if (argc==1) usage(argv[0]);

	// command-line parameters
	int op, bs, opt=0;
	int pcap_bufsize=0;
	while((op=getopt(argc, argv, "hqdp:e:b:c:r:i:t:s:o:")) != EOF){
		switch(op){
			case 'h':
				usage(argv[0]);
				break;
			case 'q':
				is_quiet = 1;
				break;
			case 'd':
				is_daemon = 1;
				break;
			case 'p':
				pid_file_name=set_string(optarg);
				break;
			case 'e':
				exec_cmd_name=set_string(optarg);
				break;
			case 'b':
				bs = atoi(optarg);
				if (bs>=0 && bs<=256) pcap_bufsize=bs;
				break;
			case 'o':
				opt= atoi(optarg);
				if (opt>=0 && opt<=2) optimization=opt;
				break;
			case 'c':
				destination=(char *)malloc(strlen(optarg)+1);
				strcpy(destination, optarg);
				break;
			case 'r':
				rule=(char *)malloc(strlen(optarg)+1);
				strcpy(rule, optarg);
				break;
			case 'i':
				interface=(char *)malloc(strlen(optarg)+1);
				strcpy(interface, optarg);
				break;
			case 't':
				bs=strtol(optarg, NULL, 10);
				if (bs>10 && bs<10000) active=bs;
				break;
			case 'x':
				bs=strtol(optarg, NULL, 10);
				if (bs>10 && bs<1000) fc_period=bs;
				break;
			case 's':
				password=(char *)malloc(strlen(optarg)+1);
				strcpy(password, optarg);
				break;
			default:
				break;
		}
	}
	setvbuf(stdout, NULL, _IOLBF, 0);

	ac = (struct agent_config*)malloc(sizeof (struct agent_config));
	ac->run=1; ac->quiet=0;
	ac->password=password;
	ac->connect_string=destination;
	ac->msgs_in=0; ac->msgs_out=0; ac->bad_pkts=0;
	ac->auth=0;
	ac->exec_cmd=NULL;

	// program start time
    gettimeofday(&ac->program_start, NULL);
    ac->start_t = ac->program_start.tv_sec;

	pcap_info = (struct ds_pcap_info*)malloc(sizeof(struct ds_pcap_info));
	pcap_info->fd=-1;
	pcap_info->interface=interface;
	pcap_info->filter=rule;
	pcap_info->is_registered=0;
	pcap_info->submit_handler = &(submit_data);
	pcap_info->enable_url_check=0;
	pcap_info->enable_dpi=0;
	pcap_info->flow_expiretime=active;
	pcap_info->pcap_bufsize=pcap_bufsize;
	pcap_info->optimization=optimization;

	if (pid_file_name)
		ac->pid_file_name=pid_file_name;
	else if (ac->pid_file_name==NULL)
		ac->pid_file_name=set_string("/var/run/n4agent.pid");

	char *host=NULL;
	int port;
	const char *x = strchr(ac->connect_string, ':');
	if (x==NULL) {
		int j=strlen(ac->connect_string);
		host=(char*)malloc(j+1);
	 	memset(host, '\0', j+1);
		strncpy(host, ac->connect_string, j);
		port=20002;
	}
	else {
		int j=x-ac->connect_string;
	 	host=(char*)malloc(j+1);
	 	memset(host, '\0', j+1);
	 	strncpy(host, ac->connect_string, j);
	 	port=atoi(x+1);
	}

	if (exec_cmd_name)
		ac->exec_cmd = exec_cmd_name;

	memset(&ac->servaddr, '\0', sizeof(ac->servaddr));
    ac->servaddr.sin_family = AF_INET;
    ac->servaddr.sin_addr.s_addr=inet_addr(host);
    ac->servaddr.sin_port=htons(port);

    ac->daemon|=is_daemon;

    ac->quiet|=is_quiet;
	if (!ac->quiet) {
        printf("n4agent (%s) is ready.\n", revision);
        printf("connect to host: %s, port %u, password is %s.\n", host, port, ac->password);
        printf("will execute '%s' as access script.\n", ac->exec_cmd?ac->exec_cmd:"NOTHING");
        printf("flow cache optimization level: %d\n", optimization);
    }

	if(ac->daemon)
		daemon(1, 0);

	FILE *pidfile = fopen(ac->pid_file_name, "wt");
	if (pidfile!=NULL) {
		fprintf(pidfile, "%u\n", getpid());
	   	fclose(pidfile);
	} else {
	   	fprintf(stderr, "Can't create pid-file: %s\n", strerror(errno));
	}

	// handle signals
	signal(SIGCHLD, (sig_t)signal_handler);
	signal(SIGPIPE, (sig_t)signal_handler);
	signal(SIGHUP,  (sig_t)signal_handler);
	signal(SIGUSR1, (sig_t)signal_handler);

	//fork executor
	if (ac->exec_cmd!=NULL) run_exec();

	// open socket
    ac->sockfd=socket(AF_INET,SOCK_DGRAM,0);
    fcntl(ac->sockfd, F_SETFL, O_NONBLOCK);
	//prepare buffers
	ac->pktlen=512;
    ac->sendbuffer=(unsigned char*)malloc(ac->pktlen);
    memset(ac->sendbuffer, '\0', ac->pktlen);

    ac->data_sendbuffer=(unsigned char*)malloc(1500);
    ac->data_lastpos=0;

	event_init();
    
	int n;
	client[0].fd=ac->sockfd;
	client[1].fd=-1; // for DS loop
	client[2].fd=STDIN_FILENO; // for DS loop
	client[3].fd=-1; //for flow expiration event

   	event_set(&client[0].ev, client[0].fd, EV_READ|EV_PERSIST, server_callback, NULL);
	event_add(&client[0].ev, NULL);

	if (!ac->daemon) {
		event_set(&client[2].ev, client[2].fd, EV_READ|EV_PERSIST, stdin_callback, NULL);
		event_add(&client[2].ev, NULL);
	}

	//flow timer
	expire_tv.tv_sec = fc_period; // every fc_period seconds we check flow cache for expired entries
	expire_tv.tv_usec = 0;

	evtimer_set(&client[3].ev, flow_expire_callback, NULL);
	evtimer_add(&client[3].ev, &expire_tv);

	/** send registration message to UDPDatasource **/
	submit_registration();
	// TODO: what if noone responds?

	event_dispatch();

	// program stops
	close(ac->sockfd);
	gettimeofday(&ac->program_end, NULL);
	ac->end_t = ac->program_end.tv_sec;
	if (!ac->quiet) printf("n4agent terminating, run time: %u seconds.\n", (unsigned)(ac->end_t - ac->start_t));
	return 0;
}
              
void server_callback (int fd, short event, void *arg) {
    struct sockaddr_in rcv_addr;
    socklen_t rcv_addr_len=sizeof(struct sockaddr_in);

	int n=recvfrom(ac->sockfd,ac->sendbuffer,512,0,(struct sockaddr *)&rcv_addr,&rcv_addr_len);
	//n4debug("udp msg type: udp_length=%d\n", n);

	ac->msgs_in++;
	if (rcv_addr_len!=sizeof(ac->servaddr)) {
		ac->bad_pkts++;
		return;
	} // wrong AF of input pkt
	if (memcmp(&rcv_addr.sin_addr,&ac->servaddr.sin_addr, sizeof(struct in_addr)) || rcv_addr.sin_port != ac->servaddr.sin_port) {
		ac->bad_pkts++;
		return;
	} // not from our server
	if (ac->sendbuffer[0]!=2) {
		ac->bad_pkts++;
		return;
	} // version must be 2
	ac->pktlen = getUShort(ac->sendbuffer, 1);
	if (ac->pktlen!=n) {
		ac->bad_pkts++;
		return;
	} // packet length
	if (getULong(ac->sendbuffer, 3)!=gethash(ac->sendbuffer, ac->password, ac->pktlen)) {
		ac->bad_pkts++;
		return;
	} // hash;

	int msg_type=ac->sendbuffer[7];

	n4debug("received msg type: %s, udp_length=%d, msg_length=%d\n", msgtype_name[msg_type], n, ac->pktlen);

	if (msg_type==MSG_SETFILTER) { // MSG_SETFILTER
		char *iface = (char*)malloc(ac->pktlen-8); memset(iface, '\0', ac->pktlen-8);
		char *rule = (char*)malloc(ac->pktlen-8); memset(rule, '\0', ac->pktlen-8);
		char *s1, *s2;
		pcap_info->enable_url_check=0; pcap_info->enable_dpi=0;

		s1 = strtok((char*)(ac->sendbuffer+8), "\t");
		if (s1==NULL) return;
		iface = set_string(s1);
		s1=strtok(NULL, "\t"); if (s1!=NULL) rule = set_string(s1);

		s1=strtok(NULL, "\t");
		if (s1!=NULL) {
			if (s1[0]=='1') pcap_info->enable_url_check=1;
			s1=strtok(NULL, "\t");
			if (s1!=NULL) if (s1[0]=='1') pcap_info->enable_dpi=1;
		}

		n4debug("iface=%s, rule=%s, URL=%d, DPI=%d\n", iface, rule, pcap_info->enable_url_check, pcap_info->enable_dpi);
		// TODO: apply pcap filter and run pcap engine
		if (strlen(iface)>1) { int k=if_nametoindex(iface); if (k>0) pcap_info->interface=iface; }
		if (strlen(rule)>1) { pcap_info->filter=rule; } else { pcap_info->filter=""; } 
		if (!ac->quiet) printf("n4agent registered at interface=%s, filter=%s\n", pcap_info->interface, pcap_info->filter);

		ds_libpcap_start(pcap_info);
		if (pcap_info->fd!=-1) {
			client[1].fd=pcap_info->fd;
			event_set(&client[1].ev, client[1].fd, EV_READ|EV_PERSIST, pcap_callback, NULL);
			event_add(&client[1].ev, NULL);
		}
		pcap_info->is_registered=1;
		ac->seq=0;
	} else if (msg_type==MSG_REGISTER) { // MSG_REGISTER (jserver got a packet from us while it was reset; asking to re-register
		pcap_info->is_registered=0;
		n4debug("server ask us to register, trying...\n");
		submit_registration();
	}
	else if (msg_type==MSG_STATUS) { // MSG_STATUS
		// no payload, we have to respond with status message
		// TODO: prepare and send response (see IPPayload.java for message structure)
		submit_keepalive();
	} else if (msg_type==MSG_AUTHACTION) { // MSG_AUTHACTION
		// see RemoteAgentDatasource.java for message structure
		unsigned char action = ac->sendbuffer[8];
		struct in_addr net, mask;
		unsigned short mbits;
		if (exec==NULL) return; // silently skip AUTHACTION
		net.s_addr = htonl(getULong(ac->sendbuffer, 9));
		mbits = getUShort(ac->sendbuffer, 13);

		int au_id = getULong(ac->sendbuffer, 17);

		char action_rule[16]; memset(action_rule, '\0', 16);
		memcpy(action_rule, ac->sendbuffer + 21, 16);

		n4debug("%s %s %s/%u au_id=%u action_rule='%s'\n", ac->exec_cmd, action_name[action], inet_ntoa(net), mbits, au_id, action_rule);
		if (action != NOTFOUND) {
			 fprintf(exec, "%s %s %s/%u %u %s\n", ac->exec_cmd, action_name[action], inet_ntoa(net), mbits, au_id, action_rule);
		}

	} // no more message types are recognized
}

void pcap_callback (int fd, short event, void *arg) {
    int status = pcap_dispatch(pcap_info->pcap, -1, ds_pcap_callback, (unsigned char *)pcap_info);
    if (status == -1) n4debug("datasource @@ pcap dispatch error: %s\n", pcap_geterr(pcap_info->pcap));                
}
              
void stdin_callback (int fd, short event, void *arg) {
    struct pcap_stat ps;
    int ch; while ((ch = getchar()) != '\n' && ch != EOF);
    if (pcap_info->is_registered) {
        printf("n4agent registered [%c%c]; in=%lu, out=%lu, bad=%lu, fc_size=%u\n", pcap_info->enable_url_check?'U':'-', pcap_info->enable_dpi?'D':'-', ac->msgs_in, ac->msgs_out, ac->bad_pkts, pcap_info->fc->entries);
        pcap_stats(pcap_info->pcap, &ps);
        printf("flowcache stats: total_entries=%llu, flushed_entries=%llu, routed_entries=%llu\n", pcap_info->fc->total_entries, pcap_info->fc->flushed_entries, pcap_info->fc->routed_entries);
        printf("pcap stats: pkt_recv=%u, pkt_drop=%u, pkt_ifdrop=%u\n", ps.ps_recv, ps.ps_drop, ps.ps_ifdrop);        
        pcap_info->fc->dump();
    } else printf("n4agent not is registered\n");
}

void flow_expire_callback (int fd, short event, void *arg) {
	if (pcap_info->fc) pcap_info->fc->expire();
	if (pcap_info->is_registered) {
		if (ac->data_lastpos > 9) forward_data_packet(); // push stalled shit out
		submit_keepalive();
	}
	else submit_registration();
	evtimer_add(&client[3].ev, &expire_tv);
}

void run_exec() {
	int fd[2];
	char buf[512];

	if (pipe(fd)) {
		n4debug("error:%s\n", strerror(errno));
	}

	if (fork()) { //parent
		close(fd[0]);
		exec = fdopen(fd[1], "w");
		setvbuf( exec, NULL, _IOLBF, 0);
		return;
	}
	//child
	exec = fdopen(fd[0], "r"); //will read from there
	close(fd[1]);

	while(!feof(exec)) {
		if (!fgets(buf, 512, exec)) {
			n4debug("error: %s\n", strerror(errno));
		}
		n4debug("exec: got %s\n", buf);
		system(buf);
	}
	exit(0);
}

/*
int cFilter        (struct cli_def *cli, const char *cmd, char **argv, int argc) {
    pcap_info->filter=set_string(argv[1]);
	return CLI_OK;
}
int cInterface      (struct cli_def *cli, const char *cmd, char **argv, int argc) {
    if (argv[1]!=NULL)  pcap_info->interface=set_string(argv[1]);
	return CLI_OK;
}
*/


